/*-
 * ========================LICENSE_START=================================
 * Copyright (C) 2021 Nordix Foundation. All rights reserved.
 * ======================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================LICENSE_END===================================
 */

package org.oransc.rappmanager.helm;

import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.oransc.rappmanager.exception.ServiceException;
import org.oransc.rappmanager.service.App;
import org.oransc.rappmanager.service.AppStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

// Supports only helm3 + version
public class HelmClient {
    private static final Logger logger = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());
    private final AppStore appStore;

    public HelmClient(AppStore appStore) {
        this.appStore = appStore;
    }

    public void installApp(App app) throws ServiceException {
        ProcessBuilder builder = prepareCreateNamespaceCommand(app.getNamespace());
        try {
            executeCommand(builder);
        } catch (ServiceException e) {
            logger.warn("Failed to create namespace {}", e.getMessage());
        }
        builder = prepareInstallCommand(app);
        executeCommand(builder);
    }

    public void uninstallApp(App app) throws ServiceException {
        ProcessBuilder builder = prepareUnInstallCommand(app);
        executeCommand(builder);
        builder = prepareDeleteNamespaceCommand(app.getNamespace());
        executeCommand(builder);
    }

    protected String getOverrideFileName(App app) {
        return appStore.getOverrideFile(app).getPath();
    }

    protected String getChartFileName(App app) {
        return appStore.getHelmChartFile(app).getPath();
    }

    protected String releaseName(App app) {
        return app.getName() + "-" + app.getVersion();
    }

    static String executeCommand(ProcessBuilder builder) throws ServiceException {
        String commandStr = toString(builder);

        try {
            Process process = builder.start();
            process.waitFor();
            int exitValue = process.exitValue();

            if (exitValue != 0) {
                String error = IOUtils.toString(process.getErrorStream(), StandardCharsets.UTF_8);
                logger.error("Executing helm command <{}> failed, exitValue: {} {}", commandStr, exitValue, error);
                throw new ServiceException("Command execution failed: " + commandStr + " " + error);
            }
            String output = IOUtils.toString(process.getInputStream(), StandardCharsets.UTF_8);
            logger.debug("Command <{}> execution, output:{}", commandStr, output);
            return output;
        } catch (Exception e) {
            logger.error("Executing helm command <{}> failed, exitValue: {}", commandStr, e.getMessage());
            throw new ServiceException("Failed to execute the Command: " + commandStr + "  " + e.getMessage());
        }
    }

    public ProcessBuilder prepareInstallCommand(App app) {

        List<String> helmArguments = Arrays.asList(//
            "helm", //
            "install", app.getName(), getChartFileName(app), //
            "--version", app.getVersion(), //
            "--namespace", app.getNamespace(), //
            "--values", getOverrideFileName(app));

        // addKubeApiInfo(helmArguments);
        ProcessBuilder processBuilder = new ProcessBuilder();
        return processBuilder.command(helmArguments);
    }

    public ProcessBuilder prepareUnInstallCommand(App app) {
        return new ProcessBuilder("helm", "delete", app.getName(), "--namespace", app.getNamespace());
    }

    // Assuming to have a separate namespace for each App,
    // modifications needed if multiple Apps could share the same namespace
    public ProcessBuilder prepareCreateNamespaceCommand(String namespace) {
        return new ProcessBuilder(Arrays.asList(//
            "kubectl", "create", "namespace", namespace));
    }

    public ProcessBuilder prepareDeleteNamespaceCommand(String namespace) {
        return new ProcessBuilder(Arrays.asList(//
            "kubectl", "delete", "namespace", namespace, "--wait=false"));
    }

    private void addKubeApiInfo(List<String> helmArguments) {
        final String KUBE_API = "--kube-apiserver";
        final String KUBE_TOKEN = "--kube-token";
        final String KUBE_CERT = "--kube-ca-file";

        List<String> kubeApiInfo = Arrays.asList(KUBE_API, "", KUBE_TOKEN, "", KUBE_CERT, "");
        helmArguments.addAll(kubeApiInfo);
    }

    protected static String toString(ProcessBuilder builder) {
        StringBuilder str = new StringBuilder();
        builder.command().forEach(arg -> str.append(" " + arg));
        return str.toString().trim();

    }

}
